{-------------------------------------------------------------------------------
 TERMS AND CONDITIONS OF USE.
 All of this software is Copyright(C) 2000 Lars Peter Christiansen.

 The author of this software assumes no liability for damages caused under
 any circumstances whatsoever, and is under no obligation. Use of the software
 indicates acceptance of all conditions contained in this document. If you do
 not agree to these terms, you must delete this software immediately.

 You may distribute the archive in which this software is distributed, but
 under no circumstances must this archive be changed. Distributing a modified
 archive is a violation of the software license.

 If you do redistribute this software, please let me know at the email address
 given below.

 If you have any questions, requests, bug reports, etc., please contact me at
 the address given below.

 Lars Peter Christiansen
 Email  : bakkevej@stofanet.dk
 Website: http://home1.stofanet.dk/nitezhifter

********************************************************************************

                                 TMappedStream

                          -By Lars Peter Christiansen

                  Encapsulates a mapped file as a stream.
                      automaticly adjusts filesize.
               Can insert, delete and replace chunks of data


********************************************************************************

-------------------------------------------------------------------------------}

unit MappedFile;

interface
uses windows,classes,sysutils;
type

  TMappedStream = Class(TCustomMemoryStream)
  private
  protected
    Ffilename,
    FobjectName:string;
    FfileHandle:Thandle;
    FmapObject : thandle;
    function OpenFile:Thandle;
    function CreateMapping(reqsize:longword):Thandle;
    function MakeView : pointer;
    procedure unmap;
    procedure SetSize(NewSize: Longint);override;
    procedure SetMapStart(const Value: longword);
  public
    Procedure FlushChanges(Amount:longword=0);
    Constructor Create(AFilename:string;ObjectName:string='');virtual;
    Destructor Destroy;override;
    function write(const Buf;count:longint):longint;override;
    Procedure Insert(Data:pointer;datasize:integer;InsertPos:longword=$FFFFFFFF);
    procedure Replace(OrgStart,OrgSize:longword;newData:pointer;newSize:longword);
    procedure Delete(DeleteSize:longword;DeleteStart:longword=$FFFFFFFF);
    property Filename:string read FFilename;
    property Handle : thandle read Ffilehandle;


  end;

implementation

{ TMappedFile }

constructor TMappedStream.Create;
begin
  inherited create;
  FFilename:=AFilename;
  FobjectName:=FobjectName;
  OpenFile;
  SetPointer(memory,GetFileSize(Ffilehandle,nil) );
  CreateMapping(size);
end;

function TMappedStream.OpenFile: Thandle;
begin
  result :=  CreateFile(pchar(FFilename),
             GENERIC_READ OR GENERIC_WRITE,
             0,
             nil,
             OPEN_ALWAYS,
             0,
             0 );
  FfileHandle:=result;
end;

function TMappedStream.CreateMapping(reqSize:longword): Thandle;
var P:Pchar;
begin
  if reqSize=0 then exit;

  if FObjectName='' then p:=nil
                    else P:=Pchar(FobjectName);

  result:= CreateFileMapping(FFileHandle,
                    nil,
                    PAGE_READWRITE,
                    0,
                    reqSize,
                    P );
  FMapObject:=result;
  if result=0 then raise
    Exception.Create('GetLastError() = '+inttostr(GetLastError));

  MakeView;
end;

procedure TMappedStream.unmap;
begin
  UnmapViewOfFile(Memory);
  CloseHandle(FmapObject);
end;

destructor TMappedStream.Destroy;
begin
  Unmap;
  CloseHandle(Ffilehandle);
  inherited;

end;

procedure TMappedStream.FlushChanges;
begin
  if not FlushViewOfFile(Memory,amount) then
      raise Exception.Create('GetLastError() = '+inttostr(GetLastError));
end;

function TMappedStream.MakeView: pointer;
begin
  result:= MapViewOfFile(FMapObject,
                FILE_MAP_ALL_ACCESS,
                0,
                0,
                0 );
  if result=nil then raise exception.create('MakeView failed');
  Setpointer(result,size);
end;





function TMappedStream.write(const Buf; count: Integer): longint;
var
  Pos: Longint;
begin
  if (Position >= 0) and (Count >= 0) then
  begin
    Pos := Position + Count;
    if Pos > 0 then
    begin
      if Pos > Size then
      begin
        SetSize( pos );
      end;
      System.Move(Buf, Pointer(Longint(Memory) + Position)^, Count);
      Position := Pos;
      Result := Count;
      Exit;
    end;
  end;
  Result := 0;
end;


procedure TMappedStream.SetSize(NewSize: Integer);
begin
  Unmap;

  if newsize<size then
    begin
      SetFilePointer(Ffilehandle,newsize,nil,FILE_BEGIN	);
      SetEndOfFile(Ffilehandle);
    end;

  SetPointer(memory,newsize);
  CreateMapping(NewSize);
end;

procedure TMappedStream.Insert;
var
  C:longword;
  Temp:pointer;
begin
  if InsertPos = $FFFFFFFF then InsertPos:=Position;

  C := size-InsertPos;             // Get size of existing data to be moved
  Size := Size + DataSize;    // Expand file to make room for new data

  GetMem(Temp,C);               // allocate temporary memory

  // copy all from right of the pos to temp.
  move(pointer(longword(memory)+InsertPos)^,
       temp^,
       C );

  // copy from temp over to new pos.
  move(temp^,
       pointer(longword(memory)+InsertPos+DataSize)^,
       C );

  // insert data at pos.
  Move(data^,
       pointer(longword(memory)+InsertPos)^,       //+0
       Datasize);                                   //10

  FreeMem(temp,C);
end;

procedure TMappedStream.Replace;
var
  Temp:pointer;
  diff:longword;
begin

  // IF THE NEW DATA IS GREATER THAN THE OLD
  if NewSize > OrgSize then  // something must be inserted
  begin
    diff := NewSize-OrgSize;

    // replace original with first part of the new data
    move(NewData^,
         pointer(longword(memory)+OrgStart)^,
         OrgSize );

    // insert the second part of the new data.
    insert(pointer(longword(newdata)+orgSize),
           diff,
           OrgStart+orgSize );

  end

  // IF THE NEW DATA IS LESS THAN THE OLD
  else if NewSize < OrgSize then // something must be deleted
    begin
      diff := OrgSize-NewSize;

      //replace first part of original with all of new data
      move(NewData^,
           Pointer(Longword(memory)+OrgStart )^,
           Newsize);

      delete(diff,
             OrgStart+NewSize );
    end

  // IF THE NEW DATA IS SAME SIZE AS THE OLD
  else
    begin
      move(NewData^,
           Pointer(Longword(memory)+OrgStart )^,
           Newsize);
    end;
end;

procedure TMappedStream.Delete;
var C:longword;
begin
  if DeleteStart=$FFFFFFFF then deleteStart:=position;

  C := size-deletesize-DeleteStart;

  Move(pointer( Longword(memory)+DeleteStart+DeleteSize)^,
       pointer( Longword(memory)+DeleteStart)^,
       C);

  Size := Size - deletesize;

end;

procedure TMappedStream.SetMapStart(const Value: longword);
begin
  CreateMapping(size);
  position:=0;
end;

end.
